/*
 *    Ensemble.java
 *    Copyright (C) 2004 Bee-Chung Chen
 *
 */

package LFA;

import weka.classifiers.Classifier;
import weka.core.OptionHandler;
import weka.core.Utils;
import weka.core.Option;
import weka.core.Instances;
import weka.core.Instance;
import java.util.Vector;
import java.util.Enumeration;
import weka.classifiers.Evaluation;

/**
 * Base class for handling ensembles.
 *
 * This class implements the Direct Ensemble method
 * of Learning From Aggregates.<p>
 *
 * Note: Class label is always the last attribute.
 *
 * Options:<p>
 *
 * -A aggregation_scheme <br>
 *    Specify the way aggregation is done. E.g. 1,2,3;2-4;1,4,5<p>
 *
 * -B classifierstring or -B classifierstring1;classifierstring2;...<br>
 *    Classifierstring, e.g. "weka.classifier.trees.J48 -U", 
 *    should contain the full class name of a scheme
 *    included for selection followed by options to the classifier. <br>
 *    "-B classifierstring" means we build the same kind of classifiers
 *    on all the training tables.<br>
 *    "-B classifierstring1;classifierstring2;..." means we build classifierstring1
 *    on the 1st training table, and so on. The number of classifierstrings should
 *    be the same as the number of training tables.<p>
 *
 * -P <br>
 *    print the base models.<p>
 *
 * -w weighting-method <br>
 *    Specify the weighting method (default: uniform).<p>
 *
 * @author Bee-Chung Chen (beechung@cs.wisc.edu)
 */
public class Ensemble extends Classifier {
  
  /** Array for storing the generated base classifiers. */
  protected Classifier[] m_Classifiers;

  /** weighting method */
  protected String m_WeightingMethod;
  
  /** Aggregation scheme */
  protected String m_AggScheme;

  /** Training tables */
  protected Instances[] m_TrainingTables;

  protected boolean printModels = false;

  /**
   * Returns an enumeration describing the available options
   *
   * -A aggregation_scheme <br>
   * Specify the way aggregation is done. E.g. 1,2,3;2-4;1,4,5<p>
   *
   * -B classifierstring or -B classifierstring1;classifierstring2;...<br>
   * Classifierstring, e.g. "weka.classifier.trees.J48 -U", 
   * should contain the full class name of a scheme
   * included for selection followed by options to the classifier. <br>
   * "-B classifierstring" means we build the same kind of classifiers
   * on all the training tables.<br>
   * "-B classifierstring1;classifierstring2;..." means we build classifierstring1
   * on the 1st training table, and so on. The number of classifierstrings should
   * be the same as the number of training tables.
   * <p>
   *
   * -w weighting-method <br>
   * Specify the weighting method (default: uniform).<p>
   *
   * @return an enumeration of all the available options
   */
  public Enumeration listOptions() {

    Vector newVector = new Vector(3);

    newVector.addElement(new Option(
	      "\tFull class name of classifier to include, followed\n"
	      + "\tby scheme options. May be specified multiple times.\n"
	      + "\t(default: \"weka.classifiers.trees.J48\")",
	      "B", 1, "-B <classifier specification>"));
    newVector.addElement(new Option(
          "\t.E.g. -A 1,2,3;2-4;1,4,5",
          "A", 1, "-A <aggregation scheme>"));
    newVector.addElement(new Option(
          "\t.E.g. -w uniform (default: uniform)",
          "w", 1, "-w <weighting method>"));
    newVector.addElement(new Option(
          "\tPrint the base models",
          "P", 0, "-P"));

    Enumeration e = super.listOptions();
    while (e.hasMoreElements()) {
      newVector.addElement(e.nextElement());
    }
    return newVector.elements();
  }

  /**
   * Parses a given list of options. Valid options are:<p>
   *
   * -A aggregation_scheme <p>
   *
   * -B classifierstring or -B classifierstring1;classifierstring2;...<p>
   *
   * -w weighting-method <p>
   *
   * @param options the list of options as an array of strings
   * @exception Exception if an option is not supported
   */
  public void setOptions(String[] options) throws Exception {

    //System.out.println("in setOptions");
    String check = Utils.getOption('c', options);
    if(check.length() > 0){
        throw new IllegalArgumentException(
                    "Class label is always the last attribute.\n"
                  ); 
    }
    
    printModels = Utils.getFlag('P', options);

    m_AggScheme = Utils.getOption('A', options);
    if(m_AggScheme.length() == 0){
        throw new IllegalArgumentException("input aggregation scheme, please.\n");        
    }
    m_AggScheme = LFA.Utils.normalizeAggScheme(m_AggScheme);
    
    String temp = Utils.getOption('w', options);
    if(temp.length() != 0) {
        m_WeightingMethod = temp;
    }else{
        m_WeightingMethod = "uniform";
    }
    
    String[] list = m_AggScheme.split(";");
    int numTables = list.length;
    m_Classifiers = new Classifier[numTables];

    // No option setting should after this part
    temp = Utils.getOption('B', options);
    if(temp.length() == 0){
        for(int i=0; i<numTables; i++){
            m_Classifiers[i] = new weka.classifiers.trees.J48();
        }
        return;
    }
    String[] strings = temp.split(";");
    if(strings.length == 1){
        String[] classifierSpec = Utils.splitOptions(strings[0]);
        String classifierName = classifierSpec[0];
        classifierSpec[0] = "";
        for(int i=0; i<numTables; i++){
            // MUST create a new array of options, because Utils.getOption(...)
            // called in forName(...) will eat options after it processed it.
            String[] opt = (String[]) classifierSpec.clone();
            m_Classifiers[i] = Classifier.forName(classifierName, opt);
        }
        return;        
    }
    if(strings.length != numTables){
        throw new IllegalArgumentException("The number of training tables and "+
                  "the number of base classifiers should be the same.\n"+
                  "Aggregation string: "+m_AggScheme+"\n"+
                  "Classifier string: "+temp+"\n");
    }
    for(int i=0; i<numTables; i++){
        String[] classifierSpec = Utils.splitOptions(strings[i]);
        String classifierName = classifierSpec[0];
        classifierSpec[0] = "";
        m_Classifiers[i] = Classifier.forName(classifierName, classifierSpec);
    }
    // No code should be after this line.
    //System.out.println("end setOptions");
  }

  /**
   * Gets the current settings of the Classifier.
   *
   * @return an array of strings suitable for passing to setOptions
   */
  public String [] getOptions() {

    // System.out.println("In getOption()");
    String [] superOptions = super.getOptions();
    int current = 0;
    String[] options = new String [superOptions.length + 6];
    if(m_AggScheme != null){
        options[current++] = "-A";
        options[current++] = m_AggScheme;
    }
    if(m_WeightingMethod != null){
        options[current++] = "-w";
        options[current++] = m_WeightingMethod;
    }
    if(m_Classifiers != null){
        StringBuffer temp = new StringBuffer();
        for(int i=0; i<m_Classifiers.length; i++){
            if(i>=1) temp.append(";");
            temp.append(getClassifierSpec(i));
        }
        options[current++] = "-B";
        options[current++] = temp.toString();
    }

    System.arraycopy(superOptions, 0, options, current, 
                     superOptions.length);
    current += superOptions.length;
    while (current < options.length) {
      options[current++] = "";
    }
    // System.out.println("End getOption(): return ");
    // for(int i=0; i<options.length; i++){
    //     System.out.println("   "+options[i]);
    // }
    return options;
  }
  
  /**
   * Returns the tip text for this property
   * @return tip text for this property suitable for
   * displaying in the explorer/experimenter gui
   */
  public String classifiersTipText() {
    return "The base classifiers to be used.";
  }

  /**
   * Sets the list of possible classifers to choose from.
   *
   * @param classifiers an array of classifiers with all options set.
   */
  public void setClassifiers(Classifier [] classifiers) {

    m_Classifiers = classifiers;
  }

  /**
   * Gets the list of possible classifers to choose from.
   *
   * @return the array of Classifiers
   */
  public Classifier [] getClassifiers() {

    return m_Classifiers;
  }
  
  /**
   * Gets a single classifier from the set of available classifiers.
   *
   * @param index the index of the classifier wanted
   * @return the Classifier
   */
  public Classifier getClassifier(int index) {
    return m_Classifiers[index];
  }
  
  public String getAggScheme(){
    return m_AggScheme;
  }
  
  /**
   * Gets the classifier specification string, which contains the class name of
   * the classifier and any options to the classifier
   *
   * @param index the index of the classifier string to retrieve, starting from
   * 0.
   * @return the classifier string, or the empty string if no classifier
   * has been assigned (or the index given is out of range).
   */
  protected String getClassifierSpec(int index) {
    
    if (m_Classifiers.length < index) {
      return "";
    }
    Classifier c = getClassifier(index);
    return c.getClass().getName() + " "
      + Utils.joinOptions(((OptionHandler)c).getOptions());
  }
  
   /**
   * Buildclassifier builds all the base classifiers
   *
   * @param data the training data.
   * @exception Exception if the classifier could not be built successfully
   */
  public void buildClassifier(Instances data) throws Exception {

    if(!LFA.Utils.isValidAggScheme(m_AggScheme,data)){
      throw new Exception("Aggregation scheme is not valid: "+m_AggScheme);
    }
    
    Instances newData = new Instances(data);
    newData.deleteWithMissingClass();

    m_TrainingTables = LFA.Utils.separateTable(newData, m_AggScheme);

    for(int i=0; i<m_Classifiers.length; i++){
      m_Classifiers[i].buildClassifier(m_TrainingTables[i]);
    }
    for(int i=0; i<m_TrainingTables.length; i++){
      m_TrainingTables[i] = new Instances(m_TrainingTables[i], 0);
    }
    //if(printModels) System.out.println(this.toString());
  }
  
  /**
   * Calculates the class membership probabilities for the given test
   * instance.
   *
   * @param instance the instance to be classified
   * @return preedicted class probability distribution
   * @exception Exception if distribution can't be computed successfully 
   */
  public double[] distributionForInstance(Instance instance) throws Exception {

    double [] sums = new double [instance.numClasses()], newProbs; 
    
    Instance[] testdata = LFA.Utils.separateInstance(instance, m_AggScheme);
    if(testdata.length != m_Classifiers.length){
      throw new Exception("Number of classifiers ("+m_Classifiers.length+") "+
                          "=! number of training tables ("+testdata.length+")");
    }
    
    for (int i = 0; i < m_Classifiers.length; i++) {
      if (instance.classAttribute().isNumeric() == true) {
        sums[0] += m_Classifiers[i].classifyInstance(testdata[i]);
      } else {
        newProbs = m_Classifiers[i].distributionForInstance(testdata[i]);
        for (int j = 0; j < newProbs.length; j++)
          sums[j] += newProbs[j];
      }
    }
    if (instance.classAttribute().isNumeric() == true) {
      sums[0] /= (double)m_Classifiers.length;
      return sums;
    } else if (Utils.eq(Utils.sum(sums), 0)) {
      return sums;
    } else {
      Utils.normalize(sums);
      return sums;
    }
  }


  /**
   * Output a representation of this classifier
   */
  public String toString() {

    if (m_Classifiers == null) {
      return "LFA.Ensemble: No model built yet.";
    }

    String result = "=== Base Classifiers ===\n\n";
    for (int i = 0; i < m_Classifiers.length; i++) {
      result += "Base Classifier "+i+": ";
      result += m_Classifiers[i].toString() + "\n";
    }

    return result;
  }

  
  /**
   * Main method for testing this class.
   *
   * @param argv should contain the following arguments:
   * -t training file [-T test file] [-c class index]
   */
  public static void main(String [] argv) {

    try {
      System.out.println(Evaluation.evaluateModel(new Ensemble(), argv));
    } catch (Exception e) {
      System.err.println(e.getMessage());
    }
  }

}
