/*
 *    OrderedBayesianEnsemble.java
 *    Copyright (C) 2004 Bee-Chung Chen
 *
 */
 
 /**
 * The Ordered Bayesian Ensemble method of
 * Learning From Aggregates.
 *
 * Options: (see also LFA.Ensemble) <p>
 * 
 * -M <number-of-random-orderings>
 *    default: the number of training tables.
 * <p>
 *
 * -D debug-level<br>
 *    1: warning, 2: possible error, 3: interesting situation, 5: detailed message
 * <p>
 *
 * @author Bee-Chung Chen (beechung@cs.wisc.edu)
 */
 
package LFA;
 
import LFA.Ensemble;
import java.util.LinkedList;
import java.util.HashMap;
import java.util.Arrays;
import java.util.Iterator;
import weka.classifiers.Classifier;
import weka.core.OptionHandler;
import weka.core.Option;
import weka.core.Instances;
import weka.core.Instance;
import java.util.Vector;
import java.util.Enumeration;
import weka.classifiers.Evaluation;
import java.io.BufferedReader;
import java.io.InputStreamReader;

public class OrderedBayesianEnsemble extends Ensemble {
  
  /** The set of aggregation schemes
   */
  protected OrderedAggScheme[] m_OrderedAggSchemes;

  /** the number of orderings */
  protected int m_NumOrderings;
  
  protected int debug = 0;

  protected double m_SmoothingFactor = 0.1;
    
  /**
   * Returns an enumeration describing the available options
   * <p>
   * Options: (see also LFA.Ensemble) <p>
   * 
   * -M <number-of-random-orderings>
   *    default: the number of training tables.
   * <p>
   *
   * -D debug-level<br>
   *    1: warning, 2: possible error, 3: interesting situation, 5: detailed message
   * <p>
   */
  public Enumeration listOptions() {

    Vector newVector = new Vector(2);

    newVector.addElement(new Option(
        "\tDefault: the number of training tables.\n",
	      "M", 1, "-M <number of random orderings>"));

    newVector.addElement(new Option(
	      "\tdebug level: 1 to 5\n",
	      "D", 1, "-D <debug level>"));

    Enumeration e = super.listOptions();
    while (e.hasMoreElements()) {
      newVector.addElement(e.nextElement());
    }
    return newVector.elements();
  }
  
  /**
   * Parses a given list of options. Valid options are:
   * <p>
   * Options: (see also LFA.Ensemble) <p>
   * 
   * -M <number-of-random-orderings>
   *    default: the number of training tables.
   * <p>
   *
   * -D debug-level<br>
   *    1: warning, 2: possible error, 3: interesting situation, 5: detailed message
   */
  public void setOptions(String[] options) throws Exception {

    String temp = weka.core.Utils.getOption('D', options);
    if(temp.length() == 0) debug = 0;
    else debug = Integer.parseInt(temp);

    temp = weka.core.Utils.getOption('M', options);
    if(temp.length() == 0) m_NumOrderings = 0;
    else m_NumOrderings = Integer.parseInt(temp);

    super.setOptions(options);
    
    generateSchemes();
  }
  
  /**
   * Buildclassifier builds all the base classifiers
   *
   * @param data the training data.
   * @exception Exception if the classifier could not be built successfully
   */
  public void buildClassifier(Instances data) throws Exception {
    super.buildClassifier(data);
  }
  

  /**
   * Calculates the class membership probabilities for the given test
   * instance.
   *
   * @param instance the instance to be classified
   * @return preedicted class probability distribution
   * @exception Exception if distribution can't be computed successfully 
   */
  public double[] distributionForInstance(Instance instance) throws Exception {

    if(instance.classAttribute().isNumeric() == true) {
    	throw new Exception("Cannot handle regression problem!!");
    }

    StringBuffer msg = new StringBuffer();

    double[] sums = new double [instance.numClasses()];
    for(int i=0; i<sums.length; i++) sums[i] = 0;
    
    Instance[] testdata = LFA.Utils.separateInstance(instance, m_AggScheme);
    if(testdata.length != m_Classifiers.length){
      throw new Exception("Number of classifiers ("+m_Classifiers.length+") "+
                          "=! number of training tables ("+testdata.length+")");
    }

    double[][] classifierProb = new double[m_Classifiers.length][];
    // classifierProb  is for estimating p(D | K - D)

    // NOTE: index i is used to loop through original training tables
    //       index m is used to loop through aggregation scheme
    for(int i = 0; i < m_Classifiers.length; i++) {
        classifierProb[i] = m_Classifiers[i].distributionForInstance(testdata[i]);
    }
    double[][] logClassifierProb = new double[classifierProb.length][];
    for(int i = 0; i<classifierProb.length; i++){
        logClassifierProb[i] = new double[classifierProb[i].length];
        //if(debug>=5) msg.append("Classifier "+i+":");
        for(int j=0; j<classifierProb[i].length; j++){
            //if(debug>=5) msg.append(" "+classifierProb[i][j]);
            if(classifierProb[i][j] == 0){
                if(debug>=2) System.out.println("Warning: probability is 0\n"+
                                 " at instance: "+instance.toString()+" ("+testdata[i]+")\n"+
                                 " classifier:  "+i+" for class: "+j);
                logClassifierProb[i][j] = LFA.Utils.SMALLEST_EXP;
            }else if(classifierProb[i][j] < 0){
                if(debug>=1) System.out.println("Warning: probability ("+classifierProb[i][j]+") less than 0\n"+
                                 " at instance: "+instance.toString()+" ("+testdata[i]+")\n"+
                                 " classifier:  "+i+" for class: "+j);
                logClassifierProb[i][j] = LFA.Utils.SMALLEST_EXP;
            }else{
                logClassifierProb[i][j] = LFA.Utils.log(classifierProb[i][j]);
            }
        }
        //if(debug>=5) msg.append("\n");
    }

    for(int k=0; k<m_OrderedAggSchemes.length; k++){

      double[] logProbs = new double[instance.numClasses()];
      double[] probs = new double[instance.numClasses()];
      int[] classifierIndex = m_OrderedAggSchemes[k].classifierIndex;
      int[][] newAttr = m_OrderedAggSchemes[k].newAttributes;
      int N = classifierIndex.length; // number of training tables in this scheme

      Instance[] testdata2 = LFA.Utils.separateInstance(
                                instance, m_OrderedAggSchemes[k].toString(), newAttr
                             );
      double[][] classifierProb2 = new double[N][];
      double[][] logClassifierProb2 = new double[N][];
      // for estimating p(D | K_old - D), index by m, not i
      for(int m=1; m<N; m++){
        int i = classifierIndex[m];
        classifierProb2[m] = m_Classifiers[i].distributionForInstance(testdata2[m]);
        logClassifierProb2[m] = new double[classifierProb2[m].length];
        int j;
        for(j=0; j<classifierProb2[m].length; j++)
            classifierProb2[m][j] += m_SmoothingFactor;
        weka.core.Utils.normalize(classifierProb2[m]);
        for(j=0; j<classifierProb2[m].length; j++){
            if(classifierProb2[m][j] == 0){
                if(debug>=2) System.out.println("Warning: probability is 0\n"+
                                 " at instance: "+instance.toString()+" ("+testdata2[m]+")\n"+
                                 " classifier:  "+i+"for class: "+j);
                logClassifierProb2[m][j] = LFA.Utils.SMALLEST_EXP;
            }else if(classifierProb2[m][j] < 0){
                if(debug>=1) System.out.println("Warning: probability ("+classifierProb2[m][j]+") less than 0\n"+
                                 " at instance: "+instance.toString()+" ("+testdata2[m]+")\n"+
                                 " classifier:  "+i+"for class: "+j);
                logClassifierProb2[m][j] = LFA.Utils.SMALLEST_EXP;
            }else{
                logClassifierProb2[m][j] = LFA.Utils.log(classifierProb2[m][j]);
            }
        }
      }

      // combine the probabilities
      for(int d=0; d<instance.numClasses(); d++){
        logProbs[d] = 0;
        int i = classifierIndex[0];
        logProbs[d] += logClassifierProb[i][d];
        if(debug>=5) msg.append("   classifier "+i+": "+LFA.Utils.exp(logClassifierProb[i][d])+"=exp"+logClassifierProb[i][d]+"\n");
        for(int m=1; m<N; m++){
          i = classifierIndex[m];
          logProbs[d] += (logClassifierProb[i][d] - logClassifierProb2[m][d]);
          if(debug>=5) msg.append("   "+testdata[i]+" / "+testdata2[m]+"\n");
          if(debug>=5) msg.append("   classifier "+i+": "+LFA.Utils.exp(logClassifierProb[i][d])+"=exp"+logClassifierProb[i][d]+"\n");
          if(debug>=5) msg.append("   classifier "+i+": "+LFA.Utils.exp(logClassifierProb2[m][d])+"=exp"+logClassifierProb2[m][d]+"\n");
        }
        if(debug>=5) msg.append("   SCORE: "+LFA.Utils.exp(logProbs[d])+" = exp("+logProbs[d]+")\n");
      }

      // adjust probabilities to prevent under/overflow
      LFA.Utils.adjustLogProbs(logProbs);
      for(int d=0; d<instance.numClasses(); d++){
        if(logProbs[d]<LFA.Utils.SMALLEST_EXP){
            if(debug>=2) System.out.println("Warning: log score ("+logProbs[d]+") is less than "+LFA.Utils.SMALLEST_EXP+"\n at instance: "+instance.toString());
            logProbs[d] = LFA.Utils.SMALLEST_EXP;
        }else if (logProbs[d]>-LFA.Utils.SMALLEST_EXP){
            if(debug>=2) System.out.println("Warning: log score ("+logProbs[d]+") is greater than "+(-LFA.Utils.SMALLEST_EXP)+"\n at instance: "+instance.toString());
            logProbs[d] = -LFA.Utils.SMALLEST_EXP;
        }
        probs[d] = LFA.Utils.exp(logProbs[d]);
        if(debug>=5) msg.append("   SCORE: "+probs[d]+" = exp("+logProbs[d]+") (AFTER Adjustment)\n");
      }

      if(! LFA.Utils.eq(weka.core.Utils.sum(probs), 0))
        weka.core.Utils.normalize(probs);
      else
        if(debug>0) System.out.println(
                       "Warning: probabilities for all classes are nearly zero at iteration "
                       +k+" (NOT Normalized)!!\n"+" at instance: "+instance.toString()
                    );
      for(int i=0; i<probs.length; i++){
      	sums[i] += probs[i];
        if(debug>=5) msg.append("FINAL "+i+": "+probs[i]+"\n");
      }
    }

    if(debug>=3){
        int maxIndex = weka.core.Utils.maxIndex(sums);
        int maxClassifierProbI[] = new int[classifierProb.length];
        for(int i=0; i<maxClassifierProbI.length; i++)
            maxClassifierProbI[i] = weka.core.Utils.maxIndex(classifierProb[i]);
        boolean same = true;
        int temp = maxClassifierProbI[0];
        for(int i=1; i<maxClassifierProbI.length; i++){
            if(maxClassifierProbI[i]!=temp){ same = false; break;}
        }
        if(same && (temp != maxIndex)){
            System.out.println("Watch Out: All classifiers agree, but the score doesn't!\n at instance: "+instance.toString());
        }
    }

    if(debug>=5) System.out.println("Instance: "+instance.toString());
    if(debug>=5) System.out.println(msg.toString());

    double total = weka.core.Utils.sum(sums);
    if(LFA.Utils.eq(total, 0)) {
      if(debug>0) System.out.println(
                     "Warning: final probability sums for all classes are nearly zero (NOT Normalized)!!\n"+
                     " at instance: "+instance.toString()
                  );
      return sums;
    } else {
      weka.core.Utils.normalize(sums);
      return sums;
    }
  }


  /**
   * Output a representation of this classifier
   */
  public String toString() {
    if (m_Classifiers == null) {
      return "LFA.OrderedBayesianEnsemble: No model built yet.";
    }
    StringBuffer result = new StringBuffer("=== Base Classifiers ===\n\n");
    for(int i=0; i<m_Classifiers.length; i++) {
      result.append("Base Classifier "+i+": ");
      result.append(m_Classifiers[i].toString() + "\n");
    }
    result.append("\nAggregation Schemes:\n");
    for(int i=0; i<m_OrderedAggSchemes.length; i++){
    	result.append("  ["+i+"]: "+m_OrderedAggSchemes[i].toString()+"\n");
        result.append("       new:"+m_OrderedAggSchemes[i].newAttrString()+"\n");
        result.append("       old:"+m_OrderedAggSchemes[i].oldAttrString()+"\n");
    }
    return result.toString();
  }

  
  /**
   * Gets the current settings of the Classifier.
   *
   * @return an array of strings suitable for passing to setOptions
   */
  public String [] getOptions() {

    String [] superOptions = super.getOptions();
    int current = 0;
    String[] options = new String [superOptions.length + 4];
    if(m_NumOrderings != 0){
        options[current++] = "-M";
        options[current++] = "" + m_NumOrderings;
    }
    if(debug > 0){
        options[current++] = "-D";
        options[current++] = "" + debug;
    }
    System.arraycopy(superOptions, 0, options, current, 
		     superOptions.length);
    current += superOptions.length;
    while (current < options.length) {
      options[current++] = "";
    }
    return options;
  }


  /**
   * Main method for testing this class.
   *
   * @param argv should contain the following arguments:
   * -t training file [-T test file] [-c class index]
   */
  public static void main(String [] argv) {

    try {
      System.out.println(Evaluation.evaluateModel(new OrderedBayesianEnsemble(), argv));
    } catch (Exception e) {
      System.err.println(e.getMessage());
      e.printStackTrace(System.err);
    }
  }

  
  private void generateSchemes() throws Exception {
    int[][] scheme = LFA.Utils.parseAggScheme(m_AggScheme);
    int N = scheme.length; // number of training tables
    if(m_NumOrderings <= 0) m_NumOrderings = N;
    int[] HasBeenChosen = new int[0];
    int numLeft = 0;
    // HasBeenChosen[k] = 1 iff table k has been chosen.
    int[] ordering = new int[N];
    int[][] newAttr = new int[N][];
    int[][] oldAttr = new int[N][];
    int numTables = 0;
    LinkedList orderedSchemeList = new LinkedList();
    for(int i=0; i<m_NumOrderings; i++){
      if( (i%N) == 0 ){
        HasBeenChosen = prepare_select(N);
        numLeft = N;
      }
      numTables = 0;
      HashMap oldHash = new HashMap(N);
      int tabIndex = random_select(numLeft, HasBeenChosen); numLeft--;
      int[][] temp = generateNewOldAttr(scheme[tabIndex], oldHash);
      ordering[numTables] = tabIndex;
      newAttr[numTables] = temp[0];
      oldAttr[numTables] = temp[1];
      numTables++;
      int[] selected = prepare_select(N);
      selected[tabIndex] = 1;
      for(int j=N-1; j>0; j--){
        tabIndex = random_select(j, selected);
        temp = generateNewOldAttr(scheme[tabIndex], oldHash);
        if(temp[0].length == 0) continue;
        ordering[numTables] = tabIndex;
        newAttr[numTables] = temp[0];
        oldAttr[numTables] = temp[1];
        numTables++;
      }
      int[] classifier_index = new int[numTables];
      int[][] new_attr = new int[numTables][];
      int[][] old_attr = new int[numTables][];
      for(int k=0; k<numTables; k++) classifier_index[k] = ordering[k];
      for(int k=0; k<numTables; k++) new_attr[k] = newAttr[k];
      for(int k=0; k<numTables; k++) old_attr[k] = oldAttr[k];
      OrderedAggScheme oas = new OrderedAggScheme(
                                   this,classifier_index,new_attr,old_attr
                                 );
      orderedSchemeList.add(oas);
    }
    Object[] temp = orderedSchemeList.toArray();
    m_OrderedAggSchemes = new OrderedAggScheme[temp.length];
    for(int i=0; i<temp.length; i++){
      m_OrderedAggSchemes[i] = (OrderedAggScheme) temp[i];
    }
  }

  // return[0]: new attributes
  // return[1]: old attributes
  private int[][] generateNewOldAttr(int[] Attrs, HashMap oldHash){
    LinkedList newAttr = new LinkedList();
    LinkedList oldAttr = new LinkedList();
    for(int i=0; i<Attrs.length; i++){
      if(oldHash.containsKey(new Integer(Attrs[i]))){
          oldAttr.add(new Integer(Attrs[i]));
      }else{
          newAttr.add(new Integer(Attrs[i]));
          oldHash.put(new Integer(Attrs[i]), null);
      }
    }
    Object[] newA =  newAttr.toArray();
    Object[] oldA =  oldAttr.toArray();
    int[][] result = new int[2][];
    result[0] = new int[newA.length];
    result[1] = new int[oldA.length];
    for(int i=0; i<newA.length; i++)
        result[0][i] = ((Integer)newA[i]).intValue();
    for(int i=0; i<oldA.length; i++)
        result[1][i] = ((Integer)oldA[i]).intValue();
    return result;
  }

  private int[] prepare_select(int num){
    int[] array = new int[num];
    for(int i=0; i<num; i++){
        array[i] = 0;
    }
    return array;
  }

  private int random_select(int numLeft, int[] selected){
    int temp = (int) (Math.random() * numLeft);
    int index = 0;
    while(selected[index]==1) index++;
    for(int i=0; i<temp; i++){
        index++;
        while(selected[index]==1) index++;
    }
    selected[index] = 1;
    return index;
  }

  /** For testing and debugging
   
  public static void main(String[] args) {
    try{
      BufferedReader input = new BufferedReader(new InputStreamReader(System.in));
      String teststring = input.readLine();
      OrderedBayesianEnsemble obe = new OrderedBayesianEnsemble();
      obe.m_AggScheme = teststring;
      obe.generateSchemes();
      for(int i=0; i<obe.m_OrderedAggSchemes.length; i++){
        System.out.println(obe.m_OrderedAggSchemes[i].toString());
        System.out.println("new attr: "+obe.m_OrderedAggSchemes[i].newAttrString());
        System.out.println("old attr: "+obe.m_OrderedAggSchemes[i].oldAttrString());
        System.out.println("");
      }
    }catch(Exception e){
      e.printStackTrace(System.out);
    }
  }
  */
}
