/*
 *    SymmetricBayesianEnsemble.java
 *    Copyright (C) 2004 Bee-Chung Chen
 *
 */
 
 /**
 * The Symmetric Bayesian Ensemble method of
 * Learning From Aggregates.
 *
 * Naive-Bayes-based Symmetric Bayesian Ensemble
 *
 * Options: (see also LFA.Ensemble) <p>
 * 
 * -M mode (default: anysub)<br>
 *    simple: every training table must have at least on 
 *            unique attritube. Otherwise, exception.
 *    direct: use the equation directly even if some table
 *            does not contain any unique attribute.
 *    anysub: use a subset of training tables in which all
 *            training tables have unique attributes to
 *            build a classifier.
 *    allsub: use all subsets of training tables in each of
 *            which all training tables have unique attributes
 *            to build several classifiers. And then combine
 *            them by uniform weighting. (support?)
 *    rand:n: use randomly chosen n subsets of training tables
 *            to build n classifiers, and then combine them
 *            by uniform weighting.
 * <p>
 *
 * -N parameter_to_Naive_Bayes<br>
 *    E.g. -N {-K} or -N {-D}<p>
 *
 * -D debug-level<br>
 *    1: warning, 2: possible error, 3: interesting situation, 5: detailed message
 * <p>
 *
 * @author Bee-Chung Chen (beechung@cs.wisc.edu)
 */
 
package LFA;
 
import LFA.NaiveBayes;
import LFA.Ensemble;
import java.util.LinkedList;
import java.util.HashMap;
import java.util.Arrays;
import java.util.Iterator;
import weka.classifiers.Classifier;
import weka.core.OptionHandler;
import weka.core.Option;
import weka.core.Instances;
import weka.core.Instance;
import java.util.Vector;
import java.util.Enumeration;
import weka.classifiers.Evaluation;


public class SymmetricBayesianEnsemble extends Ensemble {
  
  /** Naive Bayes classifier */
  protected NaiveBayes m_NaiveBayes;
  
  /** The set of aggregation schemes
   */
  protected SymmetricAggScheme[] m_SymAggSchemes;

  /** the mode */
  protected String m_Mode;
  
  /** Parameter for Naive Bayes */
  protected String m_NaiveBayesParameter;

  protected int debug = 0;
    
  /**
   * Returns an enumeration describing the available options
   * <p>
   * -M mode (default: anysub)<br>
   *    mode could be simple/direct/anysub/allsub/rand:n<br>
   *    where n is a number.
   * <p>
   * -N parameter_to_Naive_Bayes<br>
   *    E.g. -N {-K} or -N {-D}
   * <p>
   * -D debug-level
   * <p>
   */
  public Enumeration listOptions() {

    Vector newVector = new Vector(2);

    newVector.addElement(new Option(
	      "\tmode could be simple, direct, anysub, allsub or rand:n,\n"
	      + "\twhere n is a number.\n"
	      + "\t(default: anysub)",
	      "M", 1, "-M <mode>"));

    newVector.addElement(new Option(
	      "\tparameter to Naive Bayes, e.g. -N {-D} or -N -{K}\n",
	      "N", 1, "-N {<parameter to Naive Bayes>}"));

    Enumeration e = super.listOptions();
    while (e.hasMoreElements()) {
      newVector.addElement(e.nextElement());
    }
    return newVector.elements();
  }
  
  /**
   * Parses a given list of options. Valid options are:
   * <p>
   * -M mode (default: anysub)<br>
   *    mode could be simple/direct/anysub/allsub/rand:n<br>
   *    where n is a number.
   * -N parameter_to_Naive_Bayes<br>
   *    E.g. -N {-K}<p>
   */
  public void setOptions(String[] options) throws Exception {

    String temp = weka.core.Utils.getOption('D', options);
    if(temp.length() == 0) debug = 0;
    else debug = Integer.parseInt(temp);

    m_Mode = weka.core.Utils.getOption('M', options);
    if(m_Mode.length() == 0){
        m_Mode = "anysub";        
    }else{
        if(!(m_Mode.equals("anysub") || m_Mode.equals("simple") ||
             m_Mode.equals("direct") || m_Mode.equals("allsub") ) ){
    	    throw new Exception("Invalid mode: "+m_Mode);
        }
    
    }
    //System.out.println("Options: ");
    //for(int i=0; i<options.length; i++){
    //    System.out.println("    "+options[i]);
    //}
    String m_NaiveBayesParameter = weka.core.Utils.getOption('N', options);
    String param = "";
    if(m_NaiveBayesParameter.length() != 0){
        if((m_NaiveBayesParameter.charAt(0) != '{') ||
           (m_NaiveBayesParameter.charAt(m_NaiveBayesParameter.length()-1) != '}')){
            throw new Exception("Use {} after -N, e.g. -N {-D} or -N {-K}");
        }
        param = m_NaiveBayesParameter.substring(1,m_NaiveBayesParameter.length()-1);
    }
    String[] classifierSpec = weka.core.Utils.splitOptions(param);
    m_NaiveBayes = (NaiveBayes) Classifier.forName("LFA.NaiveBayes", classifierSpec);
    
    super.setOptions(options);
    
    generateSchemes(m_Mode);
  }
  
  /**
   * Buildclassifier builds all the base classifiers
   *
   * @param data the training data.
   * @exception Exception if the classifier could not be built successfully
   */
  public void buildClassifier(Instances data) throws Exception {
    super.buildClassifier(data);
    m_NaiveBayes.buildClassifier(data);
  }
  

  /**
   * Calculates the class membership probabilities for the given test
   * instance.
   *
   * @param instance the instance to be classified
   * @return preedicted class probability distribution
   * @exception Exception if distribution can't be computed successfully 
   */
  public double[] distributionForInstance(Instance instance) throws Exception {

    if(instance.classAttribute().isNumeric() == true) {
    	throw new Exception("Cannot handle regression problem!!");
    }

    StringBuffer msg = new StringBuffer();

    double[] sums = new double [instance.numClasses()];
    for(int i=0; i<sums.length; i++) sums[i] = 0;
    
    double[] logClassProb = new double[instance.numClasses()];
    for(int d=0; d<instance.numClasses(); d++)
    	logClassProb[d] = LFA.Utils.log(m_NaiveBayes.getClassPrior(d));
    
    // attrCondProb[i][j] = prob(Attribute i | Class = j, the instance)
    // for i starts from 0, j start from 0
    double[][] attrCondProb = m_NaiveBayes.getAttrConditionals(instance);
    
    double[][] logACondProb = new double[attrCondProb.length][];
    for(int i=0; i<attrCondProb.length; i++){
    	logACondProb[i] = new double[attrCondProb[i].length];
    	for(int d=0; d<attrCondProb[i].length; d++){
    	    if(attrCondProb[i][d]>0){ 
    		    logACondProb[i][d] = LFA.Utils.log(attrCondProb[i][d]);
    		}else{
    		    logACondProb[i][d] = LFA.Utils.SMALLEST_EXP;
    	    }
    	}
    }
    
    Instance[] testdata = LFA.Utils.separateInstance(instance, m_AggScheme);
    if(testdata.length != m_Classifiers.length){
      throw new Exception("Number of classifiers ("+m_Classifiers.length+") "+
                          "=! number of training tables ("+testdata.length+")");
    }

    double[][] classifierProb = new double[m_Classifiers.length][];
    for (int i = 0; i < m_Classifiers.length; i++) {
        classifierProb[i] = m_Classifiers[i].distributionForInstance(testdata[i]);
    }
    double[][] logClassifierProb = new double[classifierProb.length][];
    for(int i = 0; i<classifierProb.length; i++){
        logClassifierProb[i] = new double[classifierProb[i].length];
        //if(debug>=5) msg.append("Classifier "+i+":");
        for(int j=0; j<classifierProb[i].length; j++){
            //if(debug>=5) msg.append(" "+classifierProb[i][j]);
            if(classifierProb[i][j] == 0){
                if(debug>=2) System.out.println("Warning: probability is 0\n"+" at instance: "+instance.toString());
                logClassifierProb[i][j] = LFA.Utils.SMALLEST_EXP;
            }else if(classifierProb[i][j] < 0){
                if(debug>=1) System.out.println("Warning: probability ("+classifierProb[i][j]+") less than 0\n"+" at instance: "+instance.toString());
                logClassifierProb[i][j] = LFA.Utils.SMALLEST_EXP;
            }else{
                logClassifierProb[i][j] = LFA.Utils.log(classifierProb[i][j]);
            }
        }
        //if(debug>=5) msg.append("\n");
    }

    for(int k=0; k<m_SymAggSchemes.length; k++){
      //System.out.println("m_SymAggScheme["+k+"]="+m_SymAggSchemes[k].toString());
      //System.out.println("m_SymAggScheme["+k+"].unique="+m_SymAggSchemes[k].uniqueAttrString());
      //System.out.println("m_SymAggScheme["+k+"].common="+m_SymAggSchemes[k].commonAttrString());
      //System.out.println("m_SymAggScheme["+k+"].others="+m_SymAggSchemes[k].otherAttrString());
      double[] logProbs = new double[instance.numClasses()];
      double[] probs = new double[instance.numClasses()];
      int[] classifierIndex = m_SymAggSchemes[k].classifierIndex;
      int[] otherAttr = m_SymAggSchemes[k].otherAttributes;
      int[] v = m_SymAggSchemes[k].numAttrOccur;
      int N = classifierIndex.length;
      
      // combine the probabilities
      for(int d=0; d<instance.numClasses(); d++){
        logProbs[d] = 0;
        logProbs[d] = (1-N)*logClassProb[d];
        if(debug>=5) msg.append("Class "+d+": "+LFA.Utils.exp(logClassProb[d])+"=exp"+logClassProb[d]+" ("+(1-N)*logClassProb[d]+")\n");
        for(int m=0; m<otherAttr.length; m++){
          int i = otherAttr[m];
          logProbs[d] += (1-v[i])*logACondProb[i][d];
          if(debug>=5) msg.append("   other "+i+": "+LFA.Utils.exp(logACondProb[i][d])+"=exp"+logACondProb[i][d]+" ("+(1-v[i])*logACondProb[i][d]+")\n");
        }
        for(int m=0; m<N; m++){
          int i = classifierIndex[m];
          logProbs[d] += logClassifierProb[i][d];
          //if(debug>=5) msg.append("   "+testdata[i]+"\n");
          if(debug>=5) msg.append("   classifier "+i+": "+LFA.Utils.exp(logClassifierProb[i][d])+"=exp"+logClassifierProb[i][d]+"\n");
        }
        if(debug>=5) msg.append("   SCORE: "+LFA.Utils.exp(logProbs[d])+" = exp("+logProbs[d]+")\n");
      }
      
      // adjust probabilities to prevent under/overflow
      LFA.Utils.adjustLogProbs(logProbs);
      for(int d=0; d<instance.numClasses(); d++){
        if(logProbs[d]<LFA.Utils.SMALLEST_EXP){
            if(debug>=2) System.out.println("Warning: log score ("+logProbs[d]+") is less than "+LFA.Utils.SMALLEST_EXP+"\n at instance: "+instance.toString());
            logProbs[d] = LFA.Utils.SMALLEST_EXP;
        }else if (logProbs[d]>-LFA.Utils.SMALLEST_EXP){
            if(debug>=2) System.out.println("Warning: log score ("+logProbs[d]+") is greater than "+(-LFA.Utils.SMALLEST_EXP)+"\n at instance: "+instance.toString());
            logProbs[d] = -LFA.Utils.SMALLEST_EXP;
        }
        probs[d] = LFA.Utils.exp(logProbs[d]);
        if(debug>=5) msg.append("   SCORE: "+probs[d]+" = exp("+logProbs[d]+") (AFTER Adjustment)\n");
      }

      if(! LFA.Utils.eq(weka.core.Utils.sum(probs), 0))
        weka.core.Utils.normalize(probs);
      else
        if(debug>0) System.out.println("Warning: probabilities for all classes are nearly zero (NOT Normalized)!!\n"+" at instance: "+instance.toString());
      for(int i=0; i<probs.length; i++){
      	sums[i] += probs[i];
        if(debug>=5) msg.append("FINAL "+i+": "+probs[i]+"\n");
      }
    }

    if(debug>=3){
        int maxIndex = weka.core.Utils.maxIndex(sums);
        int maxClassifierProbI[] = new int[classifierProb.length];
        for(int i=0; i<maxClassifierProbI.length; i++)
            maxClassifierProbI[i] = weka.core.Utils.maxIndex(classifierProb[i]);
        boolean same = true;
        int temp = maxClassifierProbI[0];
        for(int i=1; i<maxClassifierProbI.length; i++){
            if(maxClassifierProbI[i]!=temp){ same = false; break;}
        }
        if(same && (temp != maxIndex)){
            System.out.println("Watch Out: All classifiers agree, but the score doesn't!\n at instance: "+instance.toString());
        }
    }

    if(debug>=5) System.out.println("Instance: "+instance.toString());
    if(debug>=5) System.out.println(msg.toString());

    double total = weka.core.Utils.sum(sums);
    if(LFA.Utils.eq(total, 0)) {
      if(debug>0) System.out.println("Warning: probability sums for all classes are nearly zero (NOT Normalized)!!\n"+" at instance: "+instance.toString());
      return sums;
    } else {
      weka.core.Utils.normalize(sums);
      return sums;
    }
  }


  /**
   * Output a representation of this classifier
   */
  public String toString() {

    if (m_Classifiers == null) {
      return "LFA.SymmetricBayesianEnsemble: No model built yet.";
    }

    StringBuffer result = new StringBuffer("=== Base Classifiers ===\n\n");
    for(int i=0; i<m_Classifiers.length; i++) {
      result.append("Base Classifier "+i+": ");
      result.append(m_Classifiers[i].toString() + "\n");
    }
    
    result.append("\nAggregation Schemes:\n");
    for(int i=0; i<m_SymAggSchemes.length; i++){
    	result.append("  ["+i+"]: "+m_SymAggSchemes[i].toString()+"\n");
        int[] otherAttr = m_SymAggSchemes[i].otherAttributes;
        result.append("       Unique:"+m_SymAggSchemes[i].uniqueAttrString()+"\n");
        result.append("       Common:"+m_SymAggSchemes[i].commonAttrString()+"\n");
        result.append("       Others:"+m_SymAggSchemes[i].otherAttrString()+"\n");
    }

    result.append("\nNaive Bayes:\n"+m_NaiveBayes.toString());

    return result.toString();
  }

  
  /**
   * Gets the current settings of the Classifier.
   *
   * @return an array of strings suitable for passing to setOptions
   */
  public String [] getOptions() {

    String [] superOptions = super.getOptions();
    int current = 0;
    String[] options = new String [superOptions.length + 4];
    if(m_Mode != null){
        options[current++] = "-M";
        options[current++] = m_Mode;
    }
    if(m_NaiveBayesParameter != null){
        options[current++] = "-N";
        options[current++] = m_NaiveBayesParameter;
    }
    System.arraycopy(superOptions, 0, options, current, 
		     superOptions.length);
    current += superOptions.length;
    while (current < options.length) {
      options[current++] = "";
    }
    return options;
  }


  /**
   * Main method for testing this class.
   *
   * @param argv should contain the following arguments:
   * -t training file [-T test file] [-c class index]
   */
  public static void main(String [] argv) {

    try {
      System.out.println(Evaluation.evaluateModel(new SymmetricBayesianEnsemble(), argv));
    } catch (Exception e) {
      System.err.println(e.getMessage());
      e.printStackTrace(System.err);
    }
  }


  // global variables used in generateSchemes(...)
  private LinkedList tempAggSchemes = null;
  int[][] originalScheme = null;
  private HashMap computed = null;
  
  private void generateSchemes(String mode) throws Exception {
    // Initialization
    originalScheme = LFA.Utils.parseAggScheme(m_AggScheme);
    tempAggSchemes = new LinkedList();
    computed = new HashMap(1000);
    int[]   selectedTableIndex = new int[0];
    int[][] selectedUniqueAttr = new int[0][];
    int[][] selectedCommonAttr = new int[0][];
    int[] tableIndexPool = new int[originalScheme.length];
    for(int i=0; i<tableIndexPool.length; i++) tableIndexPool[i] = i;
    int[] numAttrOccur = originalNumAttrOccur(originalScheme);
    
    // Call the recursive inner subroutine
    generateSchemes(selectedTableIndex, selectedUniqueAttr, selectedCommonAttr,
                    tableIndexPool, numAttrOccur, mode);
    
    // Setup the Symmetric aggregation schemes.
    // m_SymAggSchemes = (SymmetricAggScheme[]) tempAggSchemes.toArray();
    // (The above statement doesn't work)
    cleanup(tempAggSchemes);
    Object[] obj = tempAggSchemes.toArray();
    m_SymAggSchemes = new SymmetricAggScheme[obj.length];
    for(int i=0; i<obj.length; i++){
        m_SymAggSchemes[i] = (SymmetricAggScheme) obj[i];
    }
    
    // cleanup
    tempAggSchemes = null;
    originalScheme = null;
    computed = null;
  }

  // Recursive Inner Subroutine
  //   * numAttrOccur[i]: the number occurences of the attribute i (A_i)
  //     in selected tables and table pools.
  //   * return max number of selected tables
  private int generateSchemes(
        int[]   selectedTIndex,  // Indexes of selected Tables
        int[][] selectedUAttr,   // Unique attributes in the selected tables
        int[][] selectedCAttr,   // Common attributes in the selected tables
        int[] tableIndexPool, int[] numAttrOccur, String mode
  ) throws Exception {

    // Test whether this setting has been computed
    String signature = makeSignature(selectedTIndex,tableIndexPool);
    if(computed.containsKey(signature)) 
        return ((Integer)computed.get(signature)).intValue();
    // The following code cannot help!!
    // int num = noNeedToProcess(selectedTIndex,tableIndexPool);
    // if(num > 0) return num;
    
    //System.out.println("SIGNATURE: "+signature);

    int[]   selectedTableIndex = new int[originalScheme.length];
    int[][] selectedUniqueAttr = new int[originalScheme.length][];
    int[][] selectedCommonAttr = new int[originalScheme.length][];
    int numSelected = selectedTIndex.length;
    for(int i=0; i<numSelected; i++) selectedTableIndex[i] = selectedTIndex[i];
    for(int i=0; i<numSelected; i++) selectedUniqueAttr[i] = selectedUAttr[i];
    for(int i=0; i<numSelected; i++) selectedCommonAttr[i] = selectedCAttr[i];

    // Test whether every table in the pool has any unique attribute &
    // Move tables with unique attribute(s) to selectedTables
    int lastAttr = numAttrOccur.length - 1;
    int numQualifiedTables = 0;
    int[] tempOthers = new int[lastAttr];
    for(int i=0; i<tempOthers.length; i++) tempOthers[i] = 0;
    int numOthers = 0;
    for(int k=0; k<tableIndexPool.length; k++){
        int i = tableIndexPool[k];
        int[] mask = new int[originalScheme[i].length];
        int numUnique=0, numCommon=0;
        for(int j=0; j<mask.length; j++){
            int Attr = originalScheme[i][j];
            if(numAttrOccur[Attr] == 1){
                numUnique++; mask[j] = 1;
            }else if(numAttrOccur[Attr] > 1){
                numCommon++; mask[j] = 2;
                if(Attr < lastAttr && tempOthers[Attr] == 0){
                    tempOthers[Attr] = 1;
                    numOthers ++;
                }
            }else mask[j] = 0;
        }
        // If this table contains unique attribute(s),
        // move this table to selected.
        if(numUnique>0){
            numQualifiedTables++;
            selectedTableIndex[numSelected] = i;
            selectedUniqueAttr[numSelected] = new int[numUnique];
            selectedCommonAttr[numSelected] = new int[numCommon];
            int indexU=0, indexC=0;
            for(int j=0; j<mask.length; j++){
                if(mask[j]==1){
                    selectedUniqueAttr[numSelected][indexU] = originalScheme[i][j];
                    indexU++;
                }else if(mask[j]==2){
                    selectedCommonAttr[numSelected][indexC] = originalScheme[i][j];
                    indexC++;
                }
            }
            numSelected++;
            tableIndexPool[k] = -1; // indicate this slot is empty
        }
    }
    
    // compute otherAttributes
    int[] otherAttr = new int[numOthers];
    int indexO = 0;
    for(int Attr=0; Attr<tempOthers.length; Attr++){
        if(tempOthers[Attr] == 1){
            otherAttr[indexO] = Attr;
            indexO ++;
        }
    }
    
    // If every table has unique attribute(s), output & return.
    if(numQualifiedTables >= tableIndexPool.length){
      int[] classifierIndex = new int[numSelected];
      for(int k=0; k<numSelected; k++){
        classifierIndex[k] = selectedTableIndex[k];
      }
      int[][] uniqueAttr = new int[numSelected][];
      int[][] commonAttr = new int[numSelected][];
      for(int k=0; k<numSelected; k++){
        uniqueAttr[k] = selectedUniqueAttr[k];
        commonAttr[k] = selectedCommonAttr[k];
      }
      SymmetricAggScheme sas = new SymmetricAggScheme(
          this,classifierIndex,uniqueAttr,
          commonAttr,otherAttr,numAttrOccur
      );
      tempAggSchemes.add(sas);
      computed.put(signature, new Integer(numSelected));
      //System.out.println("    "+sas);
      return numSelected;
      
    }else if("simple".equals(mode)){
        
        throw new Exception("At least one table does not have any unique attribute.");
        
    }else if("direct".equals(mode)){
        
      int[] classifierIndex = new int[originalScheme.length];
      for(int k=0; k<classifierIndex.length; k++){
        classifierIndex[k] = k;
      }
      int[][] uniqueAttr = new int[originalScheme.length][];
      int[][] commonAttr = new int[originalScheme.length][];
      for(int k=0; k<numSelected; k++){
        int i = selectedTableIndex[k];
        uniqueAttr[i] = selectedUniqueAttr[k];
        commonAttr[i] = selectedCommonAttr[k];
      }
      SymmetricAggScheme sas = new SymmetricAggScheme(
          this,classifierIndex,uniqueAttr,
          commonAttr,otherAttr,numAttrOccur
      );
      tempAggSchemes.add(sas);
      return originalScheme.length;
    }
    
    // Some tables do not have unique attribute(s)
    int[] tempTableIndexPool = new int[tableIndexPool.length-numQualifiedTables];
    int index = 0;
    for(int k=0; k<tableIndexPool.length; k++){
        if(tableIndexPool[k]!=-1){ 
            tempTableIndexPool[index] = tableIndexPool[k];
            index++;
        }
    }
    int[]   tempSelectedTIndex = new int[numSelected];
    int[][] tempSelectedUAttr  = new int[numSelected][];
    int[][] tempSelectedCAttr  = new int[numSelected][];
    for(int i=0; i<numSelected; i++) tempSelectedTIndex[i] = selectedTableIndex[i];
    for(int i=0; i<numSelected; i++) tempSelectedUAttr[i]  = selectedUniqueAttr[i];
    for(int i=0; i<numSelected; i++) tempSelectedCAttr[i]  = selectedCommonAttr[i];

    int maxNumSelected = 0;

    for(int k=0; k<tempTableIndexPool.length; k++){
        int[] newTableIndexPool = new int[tempTableIndexPool.length-1];
        for(int j=0, x=0; j<tempTableIndexPool.length; j++){
            if(j!=k){
                newTableIndexPool[x] = tempTableIndexPool[j];
                x++;
            }
        }
        int[] newNumAttrOccur = (int[]) numAttrOccur.clone();
        int i = tempTableIndexPool[k];
        for(int j=0; j<originalScheme[i].length; j++){
            int Attr = originalScheme[i][j];
            newNumAttrOccur[Attr] --;
        }
        int temp = generateSchemes(
                       tempSelectedTIndex,tempSelectedUAttr,tempSelectedCAttr,
                       newTableIndexPool,newNumAttrOccur,mode
                   );
        if(temp > maxNumSelected) maxNumSelected = temp;
        if("anysub".equals(mode)) return maxNumSelected;
        if(maxNumSelected == numSelected){
            computed.put(signature, new Integer(maxNumSelected));
            return maxNumSelected;
        }
    }
    
    for(int k=0; k<numSelected; k++){
        int[]  newSelectedTIndex = new int[numSelected-1];
        int[][] newSelectedUAttr = new int[numSelected-1][];
        int[][] newSelectedCAttr = new int[numSelected-1][];
        for(int j=0, x=0; j<numSelected; j++){
            if(j!=k){
                newSelectedTIndex[x] = selectedTableIndex[j];
                newSelectedUAttr[x]  = selectedUniqueAttr[j];
                newSelectedCAttr[x]  = selectedCommonAttr[j];
                x++;
            }
        }
        int[] newTableIndexPool = (int[]) tempTableIndexPool.clone();
        int[] newNumAttrOccur = (int[]) numAttrOccur.clone();
        int i = selectedTableIndex[k];
        for(int j=0; j<originalScheme[i].length; j++){
            int Attr = originalScheme[i][j];
            newNumAttrOccur[Attr] --;
        }
        int temp = generateSchemes(
                       newSelectedTIndex,newSelectedUAttr,newSelectedCAttr,
                       newTableIndexPool,newNumAttrOccur,mode
                   );
        if(temp > maxNumSelected) maxNumSelected = temp;
    }
    
    computed.put(signature, new Integer(maxNumSelected));
    return maxNumSelected;
  }
  
  void cleanup(LinkedList tempAggSchemes){
    Iterator iter = tempAggSchemes.iterator();
    Object[] array = tempAggSchemes.toArray();
    int i=0;
    while(iter.hasNext()){
        SymmetricAggScheme scheme1 = (SymmetricAggScheme) iter.next();
        for(int j=0; j<array.length; j++){
            if(i!=j){
                SymmetricAggScheme scheme2 = (SymmetricAggScheme) array[j];
                if(LFA.Utils.isSubsetOf(scheme1.classifierIndex, scheme2.classifierIndex)){
                    iter.remove();
                    break;
                }
            }
        }
        i++;
    }
  }
  
  int noNeedToProcess(int[] selectedTableIndex, int[] tableIndexPool){
    Iterator iter = tempAggSchemes.iterator();
    while(iter.hasNext()){
        SymmetricAggScheme ascheme = (SymmetricAggScheme) iter.next();
        //System.out.println("    subset "+makeSignature(selectedTableIndex,tableIndexPool)+" ? "+LFA.Utils.toString(ascheme.classifierIndex));
        if(LFA.Utils.isSubsetOf(selectedTableIndex, ascheme.classifierIndex) &&
           LFA.Utils.isSubsetOf(tableIndexPool, ascheme.classifierIndex) ){
            return ascheme.classifierIndex.length;
        }
    }
    return -1;
  }
  
  String makeSignature(int[] selectedTableIndex, int[] tableIndexPool){
    int[] temp = new int[selectedTableIndex.length+tableIndexPool.length];
    int i=0;
    for(i=0; i<selectedTableIndex.length; i++) temp[i] = selectedTableIndex[i];
    for(int j=0; j<tableIndexPool.length; j++, i++) temp[i]= tableIndexPool[j];
    Arrays.sort(temp);
    StringBuffer str = new StringBuffer();
    for(i=0; i<temp.length; i++){
        if(i>0) str.append(",");
        str.append(temp[i]);
    }
    return str.toString();
  }
  
  private int[] originalNumAttrOccur(int[][] scheme) throws Exception {

    int lastAttr = 0;
    for(int i=0; i<scheme.length; i++){
        for(int j=0; j<scheme[i].length; j++){
            if(scheme[i][j] > lastAttr) lastAttr = scheme[i][j];
        }
    }
    
    // The result (attribute index start from 0, not 1)
    int[] AttrOccurence = new int[lastAttr+1];
    
    // used to check whether an attribute occurs in a table
    // more than once.
    // AttrTableMap[x][y] = 1 means A_x occurs in table_y
    int[][] AttrTableMap  = new int[lastAttr+1][scheme.length];
    
    for(int i=0; i<AttrTableMap.length; i++)
        for(int j=0; j<AttrTableMap[i].length; j++)
            AttrTableMap[i][j] = 0;
    for(int i=0; i<AttrOccurence.length; i++)
        AttrOccurence[i] = 0;
    
    for(int i=0; i<scheme.length; i++){
        for(int j=0; j<scheme[i].length; j++){
            int Attr = scheme[i][j];
            if(AttrTableMap[Attr][i] == 0){
                AttrTableMap[Attr][i] = 1;
                AttrOccurence[Attr]++;
            }else{
                throw new Exception(
                   "An attribute occurs in training table "+i+" twice!");
            }
        }
    }
    return AttrOccurence;
  }
  
  /** For testing and debugging
  
  public static void main(String[] args){
    String teststring = "1,2,10;2,3,10;3,4,10;2,7,10;3,6,9,10;2,8,3,10;4,1,10";
    System.out.println("INPUT: "+teststring);
    try{
      SymmetricBayesianEnsemble sbe = new SymmetricBayesianEnsemble();
      sbe.m_AggScheme = teststring;
      sbe.generateSchemes("all");
      for(int i=0; i<sbe.m_SymAggSchemes.length; i++){
        System.out.println(sbe.m_SymAggSchemes[i].toString());
      }
    }catch(Exception e){
      e.printStackTrace(System.out);
    }
  }
  */
}
