/*
 *    SymmetricBayesianEnsemble2.java
 *    Copyright (C) 2004 Bee-Chung Chen
 *
 */
 
 /**
 * The Symmetric Bayesian Ensemble method 2 of
 * Learning From Aggregates.
 *
 * It is the missing-value-based Symmetric Bayesian Ensemble.
 *
 * Options: (see also LFA.Ensemble) <p>
 * 
 * -M mode (default: anysub)<br>
 *    simple: every training table must have at least on 
 *            unique attritube. Otherwise, exception.
 *    direct: use the equation directly even if some table
 *            does not contain any unique attribute.
 *    anysub: use a subset of training tables in which all
 *            training tables have unique attributes to
 *            build a classifier.
 *    allsub: use all subsets of training tables in each of
 *            which all training tables have unique attributes
 *            to build several classifiers. And then combine
 *            them by uniform weighting. (support?)
 *    rand:n: use randomly chosen n subsets of training tables
 *            to build n classifiers, and then combine them
 *            by uniform weighting.
 * <p>
 *
 * -N parameter_to_Naive_Bayes<br>
 *    E.g. -N {-K} or -N {-D}<p>
 *
 * -D debug-level<br>
 *    1: warning, 2: possible error, 3: interesting situation, 5: detailed message
 * <p>
 *
 * @author Bee-Chung Chen (beechung@cs.wisc.edu)
 */

package LFA;
import LFA.NaiveBayes;
import LFA.Ensemble;
import weka.core.Instances;
import weka.core.Instance;
import weka.classifiers.Evaluation;

public class SymmetricBayesianEnsemble2
    extends SymmetricBayesianEnsemble {

  /**
   * Calculates the class membership probabilities for the given test
   * instance.
   *
   * @param instance the instance to be classified
   * @return preedicted class probability distribution
   * @exception Exception if distribution can't be computed successfully 
   */
  public double[] distributionForInstance(Instance instance) throws Exception {

    if(instance.classAttribute().isNumeric() == true) {
    	throw new Exception("Cannot handle regression problem!!");
    }

    StringBuffer msg = new StringBuffer();

    double[] sums = new double [instance.numClasses()];
    for(int i=0; i<sums.length; i++) sums[i] = 0;
    
    double[] logClassProb = new double[instance.numClasses()];
    for(int d=0; d<instance.numClasses(); d++)
    	logClassProb[d] = LFA.Utils.log(m_NaiveBayes.getClassPrior(d));
    
    // attrCondProb[i][j] = prob(Attribute i | Class = j, the instance)
    // for i starts from 0, j start from 0
    double[][] attrCondProb = m_NaiveBayes.getAttrConditionals(instance);
    
    double[][] logACondProb = new double[attrCondProb.length][];
    for(int i=0; i<attrCondProb.length; i++){
    	logACondProb[i] = new double[attrCondProb[i].length];
    	for(int d=0; d<attrCondProb[i].length; d++){
    	    if(attrCondProb[i][d]>0){ 
    		    logACondProb[i][d] = LFA.Utils.log(attrCondProb[i][d]);
    		}else{
    		    logACondProb[i][d] = LFA.Utils.SMALLEST_EXP;
    	    }
    	}
    }
    
    Instance[] testdata = LFA.Utils.separateInstance(instance, m_AggScheme);
    if(testdata.length != m_Classifiers.length){
      throw new Exception("Number of classifiers ("+m_Classifiers.length+") "+
                          "=! number of training tables ("+testdata.length+")");
    }

    double[][] classifierProb = new double[m_Classifiers.length][];
    // classifierProb  is for estimating p(D | K - D)

    // NOTE: index i is used to loop through original training tables
    //       index m is used to loop through aggregation scheme
    for(int i = 0; i < m_Classifiers.length; i++) {
        classifierProb[i] = m_Classifiers[i].distributionForInstance(testdata[i]);
    }
    double[][] logClassifierProb = new double[classifierProb.length][];
    for(int i = 0; i<classifierProb.length; i++){
        logClassifierProb[i] = new double[classifierProb[i].length];
        //if(debug>=5) msg.append("Classifier "+i+":");
        for(int j=0; j<classifierProb[i].length; j++){
            //if(debug>=5) msg.append(" "+classifierProb[i][j]);
            if(classifierProb[i][j] == 0){
                if(debug>=2) System.out.println("Warning: probability is 0\n"+" at instance: "+instance.toString());
                logClassifierProb[i][j] = LFA.Utils.SMALLEST_EXP;
            }else if(classifierProb[i][j] < 0){
                if(debug>=1) System.out.println("Warning: probability ("+classifierProb[i][j]+") less than 0\n"+" at instance: "+instance.toString());
                logClassifierProb[i][j] = LFA.Utils.SMALLEST_EXP;
            }else{
                logClassifierProb[i][j] = LFA.Utils.log(classifierProb[i][j]);
            }
        }
        //if(debug>=5) msg.append("\n");
    }

    for(int k=0; k<m_SymAggSchemes.length; k++){
      //System.out.println("m_SymAggScheme["+k+"]="+m_SymAggSchemes[k].toString());
      //System.out.println("m_SymAggScheme["+k+"].unique="+m_SymAggSchemes[k].uniqueAttrString());
      //System.out.println("m_SymAggScheme["+k+"].common="+m_SymAggSchemes[k].commonAttrString());
      //System.out.println("m_SymAggScheme["+k+"].others="+m_SymAggSchemes[k].otherAttrString());
      double[] logProbs = new double[instance.numClasses()];
      double[] probs = new double[instance.numClasses()];
      int[] classifierIndex = m_SymAggSchemes[k].classifierIndex;
      int[] otherAttr = m_SymAggSchemes[k].otherAttributes;
      int[][] uniqueAttr = m_SymAggSchemes[k].uniqueAttributes;
      int N = classifierIndex.length; // number of training tables in this scheme

      Instance[] testdata2 = LFA.Utils.separateInstance(
                                instance, m_SymAggSchemes[k].toString(), uniqueAttr
                             );
      double[][] classifierProb2 = new double[N][];
      double[][] logClassifierProb2 = new double[N][];
      // for estimating p(D | K_common - D), index by m, not i
      for(int m=0; m<N; m++){
        int i = classifierIndex[m];
        classifierProb2[m] = m_Classifiers[i].distributionForInstance(testdata2[m]);
        logClassifierProb2[m] = new double[classifierProb2[m].length];
        for(int j=0; j<classifierProb2[m].length; j++){
            if(classifierProb2[m][j] == 0){
                if(debug>=2) System.out.println("Warning: probability is 0\n"+" at instance: "+instance.toString());
                logClassifierProb2[m][j] = LFA.Utils.SMALLEST_EXP;
            }else if(classifierProb2[m][j] < 0){
                if(debug>=1) System.out.println("Warning: probability ("+classifierProb2[m][j]+") less than 0\n"+" at instance: "+instance.toString());
                logClassifierProb2[m][j] = LFA.Utils.SMALLEST_EXP;
            }else{
                logClassifierProb2[m][j] = LFA.Utils.log(classifierProb2[m][j]);
            }
        }
      }

      // combine the probabilities
      for(int d=0; d<instance.numClasses(); d++){
        logProbs[d] = 0;
        logProbs[d] = logClassProb[d];
        if(debug>=5) msg.append("Class "+d+": "+LFA.Utils.exp(logClassProb[d])+"=exp"+logClassProb[d]+"\n");
        for(int m=0; m<otherAttr.length; m++){
          int i = otherAttr[m];
          logProbs[d] += logACondProb[i][d];
          if(debug>=5) msg.append("   other "+i+": "+LFA.Utils.exp(logACondProb[i][d])+"=exp"+logACondProb[i][d]+"\n");
        }
        for(int m=0; m<N; m++){
          int i = classifierIndex[m];
          logProbs[d] += (logClassifierProb[i][d] - logClassifierProb2[m][d]);
          if(debug>=5) msg.append("   "+testdata[i]+" / "+testdata2[m]+"\n");
          if(debug>=5) msg.append("   classifier "+i+": "+LFA.Utils.exp(logClassifierProb[i][d])+"=exp"+logClassifierProb[i][d]+"\n");
          if(debug>=5) msg.append("   classifier "+i+": "+LFA.Utils.exp(logClassifierProb2[m][d])+"=exp"+logClassifierProb2[m][d]+"\n");
        }
        if(debug>=5) msg.append("   SCORE: "+LFA.Utils.exp(logProbs[d])+" = exp("+logProbs[d]+")\n");
      }

      // adjust probabilities to prevent under/overflow
      LFA.Utils.adjustLogProbs(logProbs);
      for(int d=0; d<instance.numClasses(); d++){
        if(logProbs[d]<LFA.Utils.SMALLEST_EXP){
            if(debug>=2) System.out.println("Warning: log score ("+logProbs[d]+") is less than "+LFA.Utils.SMALLEST_EXP+"\n at instance: "+instance.toString());
            logProbs[d] = LFA.Utils.SMALLEST_EXP;
        }else if (logProbs[d]>-LFA.Utils.SMALLEST_EXP){
            if(debug>=2) System.out.println("Warning: log score ("+logProbs[d]+") is greater than "+(-LFA.Utils.SMALLEST_EXP)+"\n at instance: "+instance.toString());
            logProbs[d] = -LFA.Utils.SMALLEST_EXP;
        }
        probs[d] = LFA.Utils.exp(logProbs[d]);
        if(debug>=5) msg.append("   SCORE: "+probs[d]+" = exp("+logProbs[d]+") (AFTER Adjustment)\n");
      }

      if(! LFA.Utils.eq(weka.core.Utils.sum(probs), 0))
        weka.core.Utils.normalize(probs);
      else
        if(debug>0) System.out.println("Warning: probabilities for all classes are nearly zero (NOT Normalized)!!\n"+" at instance: "+instance.toString());
      for(int i=0; i<probs.length; i++){
      	sums[i] += probs[i];
        if(debug>=5) msg.append("FINAL "+i+": "+probs[i]+"\n");
      }
    }

    if(debug>=3){
        int maxIndex = weka.core.Utils.maxIndex(sums);
        int maxClassifierProbI[] = new int[classifierProb.length];
        for(int i=0; i<maxClassifierProbI.length; i++)
            maxClassifierProbI[i] = weka.core.Utils.maxIndex(classifierProb[i]);
        boolean same = true;
        int temp = maxClassifierProbI[0];
        for(int i=1; i<maxClassifierProbI.length; i++){
            if(maxClassifierProbI[i]!=temp){ same = false; break;}
        }
        if(same && (temp != maxIndex)){
            System.out.println("Watch Out: All classifiers agree, but the score doesn't!\n at instance: "+instance.toString());
        }
    }

    if(debug>=5) System.out.println("Instance: "+instance.toString());
    if(debug>=5) System.out.println(msg.toString());

    double total = weka.core.Utils.sum(sums);
    if(LFA.Utils.eq(total, 0)) {
      if(debug>0) System.out.println("Warning: probability sums for all classes are nearly zero (NOT Normalized)!!\n"+" at instance: "+instance.toString());
      return sums;
    } else {
      weka.core.Utils.normalize(sums);
      return sums;
    }
  }

  /**
   * Main method for testing this class.
   *
   * @param argv should contain the following arguments:
   * -t training file [-T test file] [-c class index]
   */
  public static void main(String [] argv) {

    try {
      System.out.println(Evaluation.evaluateModel(new SymmetricBayesianEnsemble2(), argv));
    } catch (Exception e) {
      System.err.println(e.getMessage());
      e.printStackTrace(System.err);
    }
  }

}
