/*
 *    Utils.java
 *    Copyright (C) 2004 Bee-Chung Chen
 *
 */

package LFA;

import weka.core.Instance;
import weka.core.Instances;
import weka.core.Range;
import weka.filters.unsupervised.attribute.Remove;
import weka.filters.Filter;
import java.io.*;
import java.util.Arrays;
import java.util.LinkedList;
import java.util.HashMap;

public class Utils{

    public static final int SMALLEST_EXP = -100;
    public static final double SMALL_FOR_EQ = Math.exp(-500);

    /**
     * Tests if a is equal to b.
     *
     * @param a a double
     * @param b a double
     */
    public static boolean eq(double a, double b){
        return (a - b < SMALL_FOR_EQ) && (b - a < SMALL_FOR_EQ); 
    }

    public static String normalizeAggScheme(String scheme)
        throws Exception{
        int[][] schemeArray = parseAggScheme(scheme);
        StringBuffer newScheme = new StringBuffer();
        for(int i=0; i<schemeArray.length; i++){
            if(i>0) newScheme.append(";");
            newScheme.append(weka.core.Range.indicesToRangeList(schemeArray[i]));
        }
        return newScheme.toString();
    }

    /**
     *   Parse aggregation scheme into an integer array.
     *   E.g. of scheme: "1,2,3;2-4;1,4,5" => [[0,1,2],[1,2,3],[0,3,4]]
     */
    public static int[][] parseAggScheme(String scheme) 
        throws Exception{
        String list[] = scheme.split(";");
        int[][] result = new int[list.length][];
        for(int i=0; i<list.length; i++){
            Range range = new Range(list[i]);
            range.setUpper(1000);
            int[] temp1  = range.getSelection();
            Arrays.sort(temp1);
            int[] temp2 = new int[temp1.length];
            int k=0, prev=-1;
            for(int j=0; j<temp1.length; j++){
                if(temp1[j] != prev){
                    temp2[k] = temp1[j];
                    k++;
                    prev = temp1[j];
                }
            }
            result[i] = new int[k];
            for(int j=0; j<k; j++){
                result[i][j] = temp2[j];
            }
            //for(int j=0; j<result[i].length; j++){
            //    System.out.print(" "+result[i][j]);
            //}
            //System.out.println();
        }
        return result;
    }

    public static boolean isValidAggScheme(String scheme, Instances data)
        throws Exception{
        int[][] sch = parseAggScheme(scheme);
        int classIndex = data.classIndex();
        if(classIndex < 0) classIndex = data.numAttributes()-1;
        for(int i=0; i<sch.length; i++){
            if(classIndex != sch[i][sch[i].length-1]) return false;
        }
        return true;
    }

    /**
    *   Separate a training table into several tables.
    *   Each new training table has the same number of rows
    *   as the original talbe, but different subset of the
    *   columns.
    *   E.g. of scheme: "1,2,3;2-4;1,4,5"
    */
    public static Instances[] separateTable(Instances originalTable, String scheme) 
        throws Exception{
        String list[] = scheme.split(";");
        Instances[] result = new Instances[list.length];
        int classIndex = originalTable.classIndex();
        if(classIndex < 0) classIndex = originalTable.numAttributes()-1;
        for(int i=0; i<list.length; i++){
            Range range = new Range(list[i]);
            range.setUpper(1000);
            int[] selects = range.getSelection();
            Arrays.sort(selects);
            if(selects[selects.length-1] != classIndex){
                throw new Exception("Invalid Aggregation Scheme: "+scheme);
            }
            
            Remove filter = new Remove();
            filter.setInputFormat(originalTable);
            String[] options = {"-V","-R", ""};
            options[2] = list[i];
            filter.setOptions(options);
            result[i] = Filter.useFilter(originalTable, filter);
        }
        return result;
    }
 
    /**
    *   Separate a training table into several tables.
    *   Each new training table has the same number of rows
    *   as the original talbe, but different subset of the
    *   columns.
    *   E.g. of scheme: "1,2,3;2-4;1,4,5"
    */
    public static Instance[] separateInstance(
            Instance originalInst,
            String scheme
        ) throws Exception{
        String list[] = scheme.split(";");
        Instance[] result = new Instance[list.length];
        int classIndex = originalInst.dataset().classIndex();
        if(classIndex < 0) classIndex = originalInst.dataset().numAttributes()-1;
        for(int i=0; i<list.length; i++){
            Range range = new Range(list[i]);
            range.setUpper(1000);
            int[] selects = range.getSelection();
            Arrays.sort(selects);
            if(selects[selects.length-1] != classIndex){
                throw new Exception("Invalid Aggregation Scheme: "+scheme);
            }
            
            Remove filter = new Remove();
            filter.setInputFormat(originalInst.dataset());
            String[] options = {"-V","-R", ""};
            options[2] = list[i];
            filter.setOptions(options);
            filter.input(originalInst);
            filter.batchFinished();
            result[i] = filter.output();
        }
        return result;
    }
 
    /**
    *   Separate a training table into several tables.
    *   Each new training table has the same number of rows
    *   as the original talbe, but different subset of the
    *   columns.
    *   E.g. of scheme: "1,2,3;2-4;1,4,5"
    */
    public static Instance[] separateInstance(
            Instance originalInst,
            String scheme,
            int[][] missingAttr
        ) throws Exception{
        String list[] = scheme.split(";");
        if(missingAttr.length != list.length){
            throw new Exception("missing attribute list and aggregation scheme miss match\n");
        }
        Instance[] result = new Instance[list.length];
        int classIndex = originalInst.dataset().classIndex();
        if(classIndex < 0) classIndex = originalInst.dataset().numAttributes()-1;
        for(int i=0; i<list.length; i++){
            Range range = new Range(list[i]);
            range.setUpper(1000);
            int[] selects = range.getSelection();
            Arrays.sort(selects);
            if(selects[selects.length-1] != classIndex){
                throw new Exception("Invalid Aggregation Scheme: "+scheme);
            }
            Instance tempinst = new Instance(originalInst);
            for(int j=0; j<missingAttr[i].length; j++){
                tempinst.setMissing(missingAttr[i][j]);
            }
            Remove filter = new Remove();
            filter.setInputFormat(originalInst.dataset());
            String[] options = {"-V","-R", ""};
            options[2] = list[i];
            filter.setOptions(options);
            filter.input(tempinst);
            filter.batchFinished();
            result[i] = filter.output();
        }
        return result;
    }


    public static boolean belongsTo(int key, int[] array){
        for(int i=0; i<array.length; i++){
            if(array[i] == key) return true;
        }
        return false;
    }
    
    public static boolean isSubsetOf(int[] subset, int[] set){
        for(int i=0; i<subset.length; i++){
            if(!belongsTo(subset[i], set)) return false;
        }
        return true;
    }
 
    public static String toString(int[] array){
        StringBuffer str = new StringBuffer();
        for(int i=0; i<array.length; i++){
            if(i>0) str.append(",");
            str.append(array[i]);
        }
        return str.toString();
    }
 
    public static double max(double a[]){
        double temp = -(Double.MAX_VALUE);
        for(int i=0; i<a.length; i++){
            if(a[i] > temp) temp = a[i];
        }
        return temp;
    }
    
    public static void adjustLogProbs(double logProbs[]){
        double maxLogProb = max(logProbs);
        for(int i=0; i<logProbs.length; i++){
            logProbs[i] -= maxLogProb;
        }
    }
 
    public static double log(double x) throws Exception{
        if(x<0) throw new Exception("log("+x+") is not defined");
        if(x==0){
            System.err.println("Warning: log("+x+"), treated as e"+SMALLEST_EXP+"!!");
            return SMALLEST_EXP;
        }
        double temp = Math.log(x);
        if(temp <= SMALLEST_EXP) return SMALLEST_EXP;
        return temp;
    }
    
    
    public static double exp(double x) throws Exception{
        //if(x<SMALLEST_EXP) throw new Exception("TOO SMALL");
    	if(x<SMALLEST_EXP){
          System.err.println("Warning: exp("+x+"), treated as e"+SMALLEST_EXP+"!!");
    	    return Math.exp(SMALLEST_EXP);
    	}
    	return Math.exp(x);
    }
 
    /**
    *   Just for debugging
    */
    public static void main(String[] args) throws Exception{
        String testdata = "breast-w";
        String testScheme = "1,2,3;2-4;1,10";
        FileReader reader = new FileReader(testdata+".arff");
        Instances input = new Instances(reader);
        Instances[] output = separateTable(input, testScheme);
        for(int i=0; i<output.length; i++){
            FileWriter fw = new FileWriter( testdata + "."+i+".arff", false );
            PrintWriter pw = new PrintWriter( fw, true );
            pw.print(output[i].toString());
            pw.flush();
            pw.close();
        }
    }

    /*
    *   Parse aggregation scheme into an array of HashMaps.
    *   E.g. of scheme: "1,2,3;2-4;1,4,5" => [[1,2,3],[2,3,4],[1,4,5]]
    public static HashMap[] parseAggScheme(String scheme) 
        throws Exception{
        String list[] = scheme.split(";");
        HashMap[] result = new HashMap[list.length];
        for(int i=0; i<list.length; i++){
            range = new Range(list[i]);
            int[] temp = range.getSelection();
            for(int j=0 ;j<temp.length; j++){
                result[i].put(temp[j]+"", temp[j]+"");
            }
        }
        return result;
    }
    */

}
